"""
Utilities for creating equinox filtered transformations. These routines
are modified from [`zodiax`](https://github.com/LouisDesdoigts/zodiax), which
was created for the project [`dLux`](https://louisdesdoigts.github.io/dLux/).
"""

from functools import partial, wraps
from typing import Any, Callable, Hashable, Optional, Union

import equinox as eqx
from jaxtyping import PyTree


def filter_grad_with_spec(
    func: Callable,
    filter_spec: PyTree[Union[bool, Callable[[Any], bool]]],
    is_leaf: Optional[Callable[[Any], bool]] = None,
    *,
    has_aux: bool = False,
) -> Callable:
    @wraps(func)
    def partition_and_recombine_fn(pytree: PyTree, *args: Any, **kwargs: Any):
        @partial(
            eqx.filter_grad,
            has_aux=has_aux,
        )
        def recombine_fn(
            pytree_if_true: PyTree, pytree_if_false: PyTree, *args: Any, **kwargs: Any
        ):
            pytree = eqx.combine(pytree_if_true, pytree_if_false, is_leaf=is_leaf)
            return func(pytree, *args, **kwargs)

        pytree_if_true, pytree_if_false = eqx.partition(
            pytree, filter_spec, is_leaf=is_leaf
        )
        return recombine_fn(pytree_if_true, pytree_if_false, *args, **kwargs)

    return partition_and_recombine_fn


def filter_value_and_grad_with_spec(
    func: Callable,
    filter_spec: PyTree[Union[bool, Callable[[Any], bool]]],
    is_leaf: Optional[Callable[[Any], bool]] = None,
    *,
    has_aux: bool = False,
) -> Callable:
    @wraps(func)
    def partition_and_recombine_fn(pytree: PyTree, *args: Any, **kwargs: Any):
        @partial(
            eqx.filter_value_and_grad,
            has_aux=has_aux,
        )
        def recombine_fn(
            pytree_if_true: PyTree, pytree_if_false: PyTree, *args: Any, **kwargs: Any
        ):
            pytree = eqx.combine(pytree_if_true, pytree_if_false, is_leaf=is_leaf)
            return func(pytree, *args, **kwargs)

        pytree_if_true, pytree_if_false = eqx.partition(
            pytree, filter_spec, is_leaf=is_leaf
        )
        return recombine_fn(pytree_if_true, pytree_if_false, *args, **kwargs)

    return partition_and_recombine_fn


def filter_vmap_with_spec(
    func: Callable,
    filter_spec: PyTree[Union[bool, Callable[[Any], bool]]],
    is_leaf: Optional[Callable[[Any], bool]] = None,
    *,
    in_axes: PyTree[Union[int, None, Callable[[Any], Optional[int]]]] = eqx.if_array(
        axis=0
    ),
    out_axes: PyTree[Union[int, None, Callable[[Any], Optional[int]]]] = eqx.if_array(
        axis=0
    ),
    axis_name: Hashable = None,
    axis_size: Optional[int] = None,
) -> Callable:
    @wraps(func)
    def partition_and_recombine_fn(pytree: PyTree, *args: Any):
        @partial(
            eqx.filter_vmap,
            in_axes=(in_axes, None),
            out_axes=out_axes,
            axis_name=axis_name,
            axis_size=axis_size,
        )
        def recombine_fn(
            pytree_if_true_with_args: tuple[PyTree, ...],
            pytree_if_false: PyTree,
        ):
            pytree_if_true, *args = pytree_if_true_with_args

            pytree = eqx.combine(pytree_if_true, pytree_if_false, is_leaf=is_leaf)
            return func(pytree, *args)

        pytree_if_true, pytree_if_false = eqx.partition(
            pytree, filter_spec, is_leaf=is_leaf
        )
        return recombine_fn((pytree_if_true, *args), pytree_if_false)

    return partition_and_recombine_fn



from functools import partial

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.tree_util as jtu

from cryojax.simulator import DiscreteConformation, DiscreteEnsemble, Instrument


def test_conformation(potential, pose, integrator, config):
    potential = tuple([potential for _ in range(3)])
    ensemble = DiscreteEnsemble(
        potential, integrator, pose, conformation=DiscreteConformation(0)
    )
    instrument = Instrument(300.0)
    _ = ensemble.scatter_to_exit_plane(instrument, config)


def test_conformation_vmap(potential, pose, integrator, config):
    # Build Ensemble
    stacked_potential = tuple([potential for _ in range(3)])
    ensemble = DiscreteEnsemble(
        stacked_potential,
        integrator,
        pose,
        conformation=jax.vmap(lambda value: DiscreteConformation(value))(
            jnp.asarray((0, 1, 2, 1, 0))
        ),
    )
    # Setup vmap
    is_vmap = lambda x: isinstance(x, DiscreteConformation)
    to_vmap = jtu.tree_map(is_vmap, ensemble, is_leaf=is_vmap)
    vmap, novmap = eqx.partition(ensemble, to_vmap)

    @partial(jax.vmap, in_axes=[0, None, None])
    def compute_conformation_stack(vmap, novmap, config):
        ensemble = eqx.combine(vmap, novmap)
        instrument = Instrument(300.0)
        return ensemble.scatter_to_exit_plane(instrument, config)

    # Vmap over conformations
    image_stack = compute_conformation_stack(vmap, novmap, config)
    assert image_stack.shape[0] == ensemble.conformation.value.shape[0]



from ._distribution import (
    AbstractDistribution as AbstractDistribution,
    AbstractMarginalDistribution as AbstractMarginalDistribution,
)
from ._gaussian_distributions import (
    IndependentFourierGaussian as IndependentFourierGaussian,
)



"""
Base classes for image operators.
"""

from abc import abstractmethod
from typing import Any, Callable
from typing_extensions import override

import jax
import jax.numpy as jnp
from equinox import AbstractVar, field, Module, Partial
from jaxtyping import Array, Float, Inexact


class AbstractImageOperator(Module, strict=True):
    """
    The base class for image operators that contain
    model parameters and compute an ``Array`` at runtime.
    """

    @abstractmethod
    def __call__(self, *args: Any, **kwargs: Any) -> Array:
        raise NotImplementedError

    def __add__(self, other) -> "AbstractImageOperator":
        if isinstance(other, (AbstractImageOperator, Partial)):
            return SumImageOperator(self, other)
        return SumImageOperator(self, Constant(other))

    def __radd__(self, other) -> "AbstractImageOperator":
        if isinstance(other, (AbstractImageOperator, Partial)):
            return SumImageOperator(other, self)
        return SumImageOperator(Constant(other), self)

    def __sub__(self, other) -> "AbstractImageOperator":
        if isinstance(other, (AbstractImageOperator, Partial)):
            return DiffImageOperator(self, other)
        return DiffImageOperator(self, Constant(other))

    def __rsub__(self, other) -> "AbstractImageOperator":
        if isinstance(other, (AbstractImageOperator, Partial)):
            return DiffImageOperator(other, self)
        return DiffImageOperator(Constant(other), self)

    def __mul__(self, other) -> "AbstractImageOperator":
        if isinstance(other, (AbstractImageOperator, Partial)):
            return ProductImageOperator(self, other)
        return ProductImageOperator(self, Constant(other))

    def __rmul__(self, other) -> "AbstractImageOperator":
        if isinstance(other, (AbstractImageOperator, Partial)):
            return ProductImageOperator(other, self)
        return ProductImageOperator(Constant(other), self)


class AbstractImageMultiplier(Module, strict=True):
    """
    Base class for computing and applying an ``Array`` to an image.

    Attributes
    ----------
    operator :
        The operator. Note that this is automatically
        computed upon instantiation.
    """

    buffer: AbstractVar[
        Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]
    ]

    def __call__(
        self, image: Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]
    ) -> Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]:
        return image * jax.lax.stop_gradient(self.buffer)

    def __mul__(self, other) -> "AbstractImageMultiplier":
        return ProductImageMultiplier(operator1=self, operator2=other)

    def __rmul__(self, other) -> "AbstractImageMultiplier":
        return ProductImageMultiplier(operator1=other, operator2=self)


class Constant(AbstractImageOperator, strict=True):
    """An operator that is a constant."""

    value: Float[Array, "..."] = field(default=1.0, converter=jnp.asarray)

    @override
    def __call__(self, *args: Any, **kwargs: Any) -> Float[Array, ""]:
        return self.value


Constant.__init__.__doc__ = """**Arguments:**

- `value`: The value of the constant
"""


class Lambda(AbstractImageOperator, strict=True):
    """An operator that calls a custom function."""

    fn: Callable[
        [Array], Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]
    ] = field(static=True)

    @override
    def __call__(
        self, *args: Any, **kwargs: Any
    ) -> Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]:
        return self.fn(*args, **kwargs)


Lambda.__init__.__doc__ = """**Arguments:**

- `fn`: The `Callable` wrapped into a `AbstractImageOperator`.
"""


class Empirical(AbstractImageOperator, strict=True):
    """This operator stores and returns an array, rather than
    computing one from a model.
    """

    array: (
        Inexact[Array, "... y_dim x_dim"]
        | Inexact[Array, "... z_dim y_dim x_dim"]
        | Float[Array, "..."]
    )
    amplitude: Float[Array, "..."] = field(default=1.0, converter=jnp.asarray)

    @override
    def __call__(self, *args: Any, **kwargs: Any) -> Inexact[Array, "y_dim x_dim"]:
        """Return the scaled and offset measurement."""
        return self.amplitude * jax.lax.stop_gradient(self.array)


Empirical.__init__.__doc__ = """**Arguments:**

- `array`: The array to be returned upon calling `Empirical`.
- `amplitude`: An amplitude scaling for `array`.
"""


class ProductImageMultiplier(AbstractImageMultiplier, strict=True):
    """A helper to represent the product of two operators."""

    buffer: Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]

    operator1: AbstractImageMultiplier
    operator2: AbstractImageMultiplier

    def __init__(
        self,
        operator1: AbstractImageMultiplier,
        operator2: AbstractImageMultiplier,
    ):
        self.operator1 = operator1
        self.operator2 = operator2
        self.buffer = operator1.buffer * operator2.buffer

    def __repr__(self):
        return f"{repr(self.operator1)} * {repr(self.operator2)}"


class SumImageOperator(AbstractImageOperator, strict=True):
    """A helper to represent the sum of two operators."""

    operator1: AbstractImageOperator | Partial
    operator2: AbstractImageOperator | Partial

    @override
    def __call__(self, *args: Any, **kwargs: Any) -> Array:
        return self.operator1(*args, **kwargs) + self.operator2(*args, **kwargs)

    def __repr__(self):
        return f"{repr(self.operator1)} + {repr(self.operator2)}"


class DiffImageOperator(AbstractImageOperator, strict=True):
    """A helper to represent the difference of two operators."""

    operator1: AbstractImageOperator | Partial
    operator2: AbstractImageOperator | Partial

    @override
    def __call__(self, *args: Any, **kwargs: Any) -> Array:
        return self.operator1(*args, **kwargs) - self.operator2(*args, **kwargs)

    def __repr__(self):
        return f"{repr(self.operator1)} - {repr(self.operator2)}"


class ProductImageOperator(AbstractImageOperator, strict=True):
    """A helper to represent the product of two operators."""

    operator1: AbstractImageOperator | Partial
    operator2: AbstractImageOperator | Partial

    @override
    def __call__(self, *args: Any, **kwargs: Any) -> Array:
        return self.operator1(*args, **kwargs) * self.operator2(*args, **kwargs)

    def __repr__(self):
        return f"{repr(self.operator1)} * {repr(self.operator2)}"
